import { CompileTarget } from "../compile";
import { TypeDefinition, EnumDefinition, ServiceDefinition, Step } from "../ir";

type lineAppender = (ind: number, line: string | string[]) => void;

const conversion = {
   boolean: "bool",
   float: "f64",
   int: "i64",
   string: "String",
   void: "void",
   bytes: "Vec<u8>",
};

function toRustType(type: string): string {
   return (conversion as any)[type] || type;
}

function toSnake(input: string) {
   return (
      input[0].toLowerCase() +
      input.slice(1).replace(/[A-Z]/g, (letter) => `_${letter.toLowerCase()}`)
   );
}

export class RustTarget extends CompileTarget<{ rust_crate: string }> {
   name: string = "rust";

   get crate() {
      return this.options.rust_crate;
   }

   start(): void {
      if (!this.crate)
         throw new Error(
            "Setting a crate name is required. Add the following to your jrpc file: 'define rust_crate <name>'"
         );

      if (this.options.allow_bytes == true) {
         throw new Error("Rust has no support for 'bytes' yet!");
      }

      this.writeFile(
         "Cargo.toml",
         this.getTemplate("Rust/Cargo.toml").replace("__name__", this.crate)
      );
   }

   private addDependencies(
      a: lineAppender,
      def: TypeDefinition | ServiceDefinition
   ) {
      for (const dep of def.depends) {
         a(0, `use crate::${dep};`);
      }

      a(0, ``);
      a(0, ``);
   }

   generateType(definition: TypeDefinition): void {
      let lines: string[] = [];
      const a: lineAppender = (i, t) => {
         if (!Array.isArray(t)) {
            t = [t];
         }
         t.forEach((l) => lines.push("   ".repeat(i) + l.trim()));
      };
      if (definition.fields.find((e) => e.map))
         a(0, `use std::collections::hash_map::HashMap;`);
      a(0, `use serde::{Deserialize, Serialize};`);
      this.addDependencies(a, definition);

      a(0, `#[derive(Clone, Debug, Serialize, Deserialize)]`);
      a(0, `pub struct ${definition.name} {`);
      for (const field of definition.fields) {
         if (field.array) {
            a(1, `pub ${field.name}: Vec<${toRustType(field.type)}>,`);
         } else if (field.map) {
            a(
               1,
               `pub ${field.name}: HashMap<${toRustType(
                  field.map
               )}, ${toRustType(field.type)}>,`
            );
         } else {
            a(1, `pub ${field.name}: ${toRustType(field.type)},`);
         }
      }
      a(0, `}`);

      this.writeFile(`src/${toSnake(definition.name)}.rs`, lines.join("\n"));
   }

   generateEnum(definition: EnumDefinition): void {
      let lines: string[] = [];
      const a: lineAppender = (i, t) => {
         if (!Array.isArray(t)) {
            t = [t];
         }
         t.forEach((l) => lines.push("   ".repeat(i) + l.trim()));
      };

      a(0, `use int_enum::IntEnum;`);
      a(0, `use serde::{Deserialize, Serialize};`);

      a(0, ``);
      a(0, ``);

      a(0, `#[repr(i64)]`);
      a(
         0,
         "#[derive(Clone, Copy, Debug, Eq, PartialEq, IntEnum, Deserialize, Serialize)]"
      );
      a(0, `pub enum ${definition.name} {`);
      for (const field of definition.values) {
         a(1, `${field.name} = ${field.value},`);
      }
      a(0, `}`);

      this.writeFile(`src/${toSnake(definition.name)}.rs`, lines.join("\n"));
   }

   private generateServiceReqResTypes(definition: ServiceDefinition) {
      let lines: string[] = [];
      const a: lineAppender = (i, t) => {
         if (!Array.isArray(t)) {
            t = [t];
         }
         t.forEach((l) => lines.push("   ".repeat(i) + l.trim()));
      };

      const paramToRust = (type: string, array: boolean) => {
         let rt = toRustType(type);
         return array ? `Vec<${rt}>` : rt;
      };

      this.addDependencies(a, definition);
      let baseName = definition.name;
      for (const fnc of definition.functions) {
         let name = definition.name + "_" + toSnake(fnc.name);
         a(
            0,
            `struct ${name}_Params(${fnc.inputs
               .map((i) => paramToRust(i.type, i.array))
               .join(",")});`
         );
         a(0, ``);
         a(0, ``);
         a(0, ``);
         a(0, ``);
         a(0, ``);
      }
      this.writeFile(
         `src/${toSnake(definition.name)}_types.rs`,
         lines.join("\n")
      );
   }

   private generateServiceClient(definition: ServiceDefinition): void {}

   private generateServiceServer(definition: ServiceDefinition): void {
      let lines: string[] = [];
      const a: lineAppender = (i, t) => {
         if (!Array.isArray(t)) {
            t = [t];
         }
         t.forEach((l) => lines.push("   ".repeat(i) + l.trim()));
      };

      this.addDependencies(a, definition);

      const retToRust = (type: string, array: boolean) => {
         let rt = toRustType(type);
         return array ? `Vec<${rt}>` : rt;
      };

      a(0, ``);
      a(0, `struct ${definition.name}ServiceHandler {`);
      a(1, `service: impl ${definition.name}`);
      a(1, ``);
      a(1, ``);
      a(0, `}`);
      a(0, ``);
      a(0, `impl ${definition.name}ServiceHandler {`);
      a(
         1,
         `fn new(service: impl ${definition.name}) -> ${definition.name}ServiceHandler {`
      );
      a(2, ``);
      a(2, ``);
      a(2, ``);
      a(1, `}`);
      a(0, `}`);
      a(0, ``);
      a(0, `trait ${definition.name} {`);

      for (const fnc of definition.functions) {
         let returnType = !fnc.return
            ? "()"
            : retToRust(fnc.return.type, fnc.return.array);
         a(1, `pub async fn ${fnc.name}(&self) -> Result<${returnType}>;`);
         // a(0, ``);
      }

      a(0, `}`);
      this.writeFile(
         `src/${toSnake(definition.name)}_types.rs`,
         lines.join("\n")
      );
   }

   generateService(definition: ServiceDefinition): void {
      this.generateServiceServer(definition);
      // throw new Error("Service not implemented.");
   }

   private generateLib(steps: Step[]) {
      let lines: string[] = [];
      const a: lineAppender = (i, t) => {
         if (!Array.isArray(t)) {
            t = [t];
         }
         t.forEach((l) => lines.push("   ".repeat(i) + l.trim()));
      };

      for (const [typ, def] of steps) {
         if (typ == "type" || typ == "enum") {
            a(0, `mod ${toSnake(def.name)};`);
            a(0, `pub use ${toSnake(def.name)}::${def.name};`);
         }
      }

      this.writeFile(`src/lib.rs`, lines.join("\n"));
   }

   finalize(steps: Step[]): void {
      this.generateLib(steps);
      // throw new Error("Method not implemented.");
   }
}
